import firebase_admin
from fastapi import FastAPI, Request
from fastapi_simple_cache import FastAPISimpleCache
from fastapi_simple_cache.backends.firestore import FirestoreBackend
from fastapi_simple_cache.decorator import cache
from firebase_admin import firestore, credentials

app = FastAPI()


@app.on_event("startup")
async def startup():
    cred = credentials.ApplicationDefault()
    firebase_admin.initialize_app(cred, {"projectId": "gcp_project"})
    db = firestore.client()
    collection = db.collection("cache_collection")
    backend = FirestoreBackend(collection=collection)
    FastAPISimpleCache.init(backend=backend)


@app.post("/")
@cache(expire=10)
def root(message: str, request: Request):
    return {"message": message}



import inspect
import logging
from fastapi.encoders import jsonable_encoder
from fastapi.requests import Request
from fastapi.responses import Response, JSONResponse
from functools import wraps
from typing import List

from .. import FastAPISimpleCache
from .key_builder import build_key

logger = logging.getLogger(__name__)


def cache(
    expire: int = 3600,
    status_codes: List[int] = [200],
):
    def wrapper(func):

        anno = func.__annotations__
        desc = f"{func.__module__}.{func.__name__}"
        request_key = next((k for k, v in anno.items() if v == Request), None)
        if request_key is None:
            raise TypeError(f"No Request parameter in {desc}")

        @wraps(func)
        async def inner(*args, **kwargs):
            nonlocal expire
            nonlocal status_codes
            nonlocal request_key
            # Get incoming request
            request = kwargs.get(request_key)
            no_cache = "no-cache" in request.headers.get("cache-control", "")
            # Get info from request
            query_params = dict(request.query_params)
            path_params = dict(request.path_params)
            body = await request.body()
            # Get response from cache
            key = build_key(
                func=func,
                namespace=FastAPISimpleCache.namespace,
                query_params=query_params,
                path_params=path_params,
                body=body,
            )
            if not no_cache:
                res, ttl = await FastAPISimpleCache.get(key=key)
                # Return if valid response exists
                if res and (ttl > 0):
                    res.headers["cache-control"] = f"max-age={expire}"
                    res.headers["age"] = f"{expire - ttl}"
                    return res
            # Get response
            if inspect.iscoroutinefunction(func):
                res = await func(**kwargs)
            else:
                res = func(**kwargs)
            # Return response
            if not isinstance(res, Response):
                res = JSONResponse(content=jsonable_encoder(res))
            if res.status_code in status_codes:
                FastAPISimpleCache.set(key, res, expire)
                if no_cache:
                    res.headers["cache-control"] = "no-cache"
                    logger.warning("Not cached: 'no-cache' directive")
                else:
                    res.headers["cache-control"] = f"max-age={expire}"
                    res.headers["age"] = "0"
            else:
                logger.warning(f"Not cached: status code {res.status_code}")
            return res

        return inner

    return wrapper



import pytest
from fastapi import FastAPI, Request
from fastapi.responses import JSONResponse
from fastapi.testclient import TestClient
from pydantic import BaseModel

from fastapi_simple_cache import FastAPISimpleCache
from fastapi_simple_cache.backends.inmemory import InMemoryBackend
from fastapi_simple_cache.decorator import cache

app = FastAPI()
client = TestClient(app)


@pytest.fixture
def backend():
    backend = InMemoryBackend()
    FastAPISimpleCache.reset()
    FastAPISimpleCache.init(backend=backend)


def test_status_code(backend, caplog):
    @app.get("/status_code")
    @cache(expire=1, status_codes=[200, 201])
    def status_code(status_code: int, request: Request):
        return JSONResponse(
            content={"status_code": status_code}, status_code=status_code
        )

    response = client.get("/status_code", params={"status_code": 404})
    assert caplog.records[0].message == "Not cached: status code 404"
    assert response.json().get("status_code") == 404
    response = client.get("/status_code", params={"status_code": 201})
    assert response.json().get("status_code") == 201
    assert response.headers.get("age") == "0"


def test_sync_get(backend):
    @app.get("/sync_get")
    @cache(expire=1)
    def sync_get(message: str, request: Request):
        return {"message": message}

    response = client.get("/sync_get", params={"message": "sync_get"})
    assert response.json().get("message") == "sync_get"


def test_async_get(backend):
    @app.get("/async_get")
    @cache(expire=1)
    async def async_get(message: str, request: Request):
        return {"message": message}

    response = client.get("/async_get", params={"message": "async_get"})
    assert response.json().get("message") == "async_get"


class RequestTest(BaseModel):
    message: str


def test_sync_post(backend):
    @app.post("/sync_post")
    @cache(expire=1)
    def sync_post(message: str, request: Request):
        return {"message": message}

    response = client.post("/sync_post", params={"message": "sync_post"})
    assert response.json().get("message") == "sync_post"


def test_async_post(backend):
    @app.post("/async_post")
    @cache(expire=1)
    async def async_post(message: RequestTest, request: Request):
        return {"message": message.message}

    response = client.post("/async_post", json={"message": "async_post"})
    assert response.json().get("message") == "async_post"



import abc
import logging
from fastapi import Response
from typing import Optional, Tuple

logger = logging.getLogger(__name__)


class Backend:
    @abc.abstractmethod
    async def get(
        self,
        key: str,
    ) -> Tuple[Optional[Response], Optional[int]]:
        raise NotImplementedError

    @abc.abstractmethod
    async def set(
        self,
        key: str,
        response: Response,
        expire: int,
    ) -> None:
        raise NotImplementedError



import json
import hashlib
from fastapi.encoders import jsonable_encoder
from typing import Callable


def build_key(
    func: Callable,
    *args,
    **kwargs,
):
    key_dict = {
        "__module__": func.__module__,
        "__name__": func.__name__,
        "__args__": args,
        "__kwargs__": jsonable_encoder(kwargs),
    }
    str2hash = json.dumps(jsonable_encoder(key_dict))
    key = hashlib.sha256(str2hash.encode())
    return key.hexdigest()



import pytest
from fastapi import FastAPI
from fastapi.testclient import TestClient

from fastapi_simple_cache.backends import Backend
from fastapi_simple_cache.decorator import cache
from fastapi_simple_cache import FastAPISimpleCache

app = FastAPI()
client = TestClient(app)


def test_no_request_param():
    with pytest.raises(TypeError) as e:

        @app.post("/no_request_param")
        @cache(expire=1)
        async def async_post(message: str):
            return {"message": message}

        assert e.value.startswith("No Request parameter in")


@pytest.mark.asyncio
async def test_no_init_get():
    FastAPISimpleCache.reset()
    with pytest.raises(Exception) as e:
        await FastAPISimpleCache.get(key="test")
        assert e.value == "You must call init first"


@pytest.mark.asyncio
async def test_no_init_set():
    FastAPISimpleCache.reset()
    with pytest.raises(Exception) as e:
        await Backend().set(key="test", response="test", expire=1)
        assert e.value == "You must call init first"


@pytest.mark.asyncio
async def test_abs_backend_get():
    with pytest.raises(NotImplementedError):
        await Backend().get(key="test")


@pytest.mark.asyncio
async def test_abs_backend_set():
    with pytest.raises(NotImplementedError):
        await Backend().set(key="test", response="test", expire=1)



import logging
from datetime import datetime, timedelta
from fastapi import Response
from google.cloud.firestore_v1.collection import CollectionReference
from math import ceil
from typing import Optional, Tuple

from . import Backend

logger = logging.getLogger(__name__)


class FirestoreBackend(Backend):
    def __init__(
        self,
        collection: CollectionReference,
    ):
        self.collection = collection

    async def get(
        self,
        key: str,
    ) -> Tuple[Optional[Response], Optional[int]]:
        doc = self.collection.document(key).get()
        if not doc.exists:
            return None, None
        data = doc.to_dict()
        response = Response(content=data.get("body"), headers=data.get("headers"))
        expire_at = data.get("expire_at").replace(tzinfo=None)
        ttl = (expire_at - datetime.utcnow()).total_seconds()
        if ttl > 0:
            logger.debug(f"Get {key} from {type(self).__name__}")
        else:
            logger.debug(f"Exp {key} from {type(self).__name__}")
        return response, ceil(ttl)

    async def set(
        self,
        key: str,
        response: Response,
        expire: int,
    ) -> None:
        body = response.body
        headers = dict(response.headers)
        expire_at = datetime.utcnow() + timedelta(seconds=expire)
        data = {"body": body, "headers": headers, "expire_at": expire_at}
        self.collection.document(key).set(data)
        logger.debug(f"Set {key} to {type(self).__name__}")



import pytest
import time
from fastapi import FastAPI, Request
from fastapi.testclient import TestClient

from fastapi_simple_cache.decorator import cache
from fastapi_simple_cache import FastAPISimpleCache
from fastapi_simple_cache.backends.inmemory import InMemoryBackend

app = FastAPI()
client = TestClient(app)


@app.get("/")
@cache(expire=2)
def root(request: Request):
    return {"time": time.time()}


@pytest.fixture
def backend():
    backend = InMemoryBackend()
    FastAPISimpleCache.reset()
    FastAPISimpleCache.init(backend=backend)


def test_max_age(backend):
    response = client.get("/")
    first_time = response.json().get("time")
    assert response.headers.get("cache-control") == "max-age=2"
    assert response.headers.get("age") == "0"
    time.sleep(1)
    response = client.get("/")
    second_time = response.json().get("time")
    assert first_time == second_time
    assert response.headers.get("cache-control") == "max-age=2"
    assert response.headers.get("age") == "1"
    time.sleep(1)
    response = client.get("/")
    third_time = response.json().get("time")
    assert first_time < third_time - 2
    assert response.headers.get("cache-control") == "max-age=2"
    assert response.headers.get("age") == "0"


def test_no_cache(backend):
    response = client.get("/", headers={"cache-control": "no-cache"})
    assert response.headers.get("cache-control") == "no-cache"



from typing import Mapping, Optional, Any


class Document:
    def __init__(
        self,
        data: Optional[Mapping] = None,
        collection: Optional[Any] = None,
        key: Optional[str] = None,
    ):
        if data is None:
            data = {}
        self.data = data
        self.collection = collection
        self.key = key

    def get(
        self,
        key: Optional[str] = None,
    ):
        if key is None:
            return self
        return self.data.get(key)

    def set(
        self,
        data: Mapping,
    ):
        self.collection.data[self.key] = data

    def to_dict(self):
        return self.data

    @property
    def exists(self):
        return len(self.data) > 0


class CollectionReference:
    def __init__(
        self,
        data: Optional[Mapping] = None,
    ):
        if data is None:
            data = {}
        self.data = data

    def document(self, key: str) -> Document:
        return Document(
            data=self.data.get(key),
            collection=self,
            key=key,
        )



import json
import logging
from datetime import datetime, timedelta
from fastapi import Response
from fastapi.encoders import jsonable_encoder
from math import ceil
from redis.asyncio.client import Redis
from typing import Optional, Tuple

from . import Backend

logger = logging.getLogger(__name__)


class RedisBackend(Backend):
    def __init__(
        self,
        redis: Redis,
    ):
        self.redis = redis

    async def get(
        self,
        key: str,
    ) -> Tuple[Optional[Response], Optional[int]]:
        async with self.redis.pipeline(transaction=True) as pipe:
            ttl, value = await (pipe.ttl(key).get(key).execute())
        if value is None:
            return None, None
        data = json.loads(value)
        response = Response(content=data.get("body"), headers=data.get("headers"))
        expire_at = datetime.fromisoformat(data.get("expire_at"))
        ttl = (expire_at - datetime.utcnow()).total_seconds()
        if ttl > 0:
            logger.debug(f"Get {key} from {type(self).__name__}")
        else:
            logger.debug(f"Exp {key} from {type(self).__name__}")
        return response, ceil(ttl)

    async def set(
        self,
        key: str,
        response: Response,
        expire: int,
    ) -> None:
        body = response.body
        headers = dict(response.headers)
        expire_at = datetime.utcnow() + timedelta(seconds=expire)
        data = {"body": body, "headers": headers, "expire_at": expire_at}
        value = json.dumps(jsonable_encoder(data), ensure_ascii=False)
        await self.redis.set(key, value, ex=expire)
        logger.debug(f"Set {key} to {type(self).__name__}")



import logging
from datetime import datetime, timedelta
from fastapi import Response
from math import ceil
from typing import Optional, Tuple

from . import Backend

logger = logging.getLogger(__name__)


class InMemoryBackend(Backend):
    def __init__(self):
        self.data = {}

    async def get(
        self,
        key: str,
    ) -> Tuple[Optional[Response], Optional[int]]:
        data = self.data.get(key)
        if data is None:
            return None, None
        response = Response(content=data.get("body"), headers=data.get("headers"))
        ttl = (data.get("expire_at") - datetime.utcnow()).total_seconds()
        if ttl > 0:
            logger.debug(f"Get {key} from {type(self).__name__}")
        else:
            logger.debug(f"Exp {key} from {type(self).__name__}")
        return response, ceil(ttl)

    async def set(
        self,
        key: str,
        response: Response,
        expire: int,
    ) -> None:
        body = response.body
        headers = dict(response.headers)
        expire_at = datetime.utcnow() + timedelta(seconds=expire)
        self.data[key] = {"body": body, "headers": headers, "expire_at": expire_at}
        logger.debug(f"Set {key} to {type(self).__name__}")



from datetime import datetime
from fastapi import FastAPI, Request
from fastapi_simple_cache import FastAPISimpleCache
from fastapi_simple_cache.backends.inmemory import InMemoryBackend
from fastapi_simple_cache.decorator import cache

app = FastAPI()


@app.on_event("startup")
async def startup():
    backend = InMemoryBackend()
    FastAPISimpleCache.init(backend=backend)


@app.get("/")
@cache(expire=10)
def root(request: Request):
    return {"datetime": datetime.utcnow()}



import pytest
import re
import time
from fastapi import FastAPI, Request
from fastapi.testclient import TestClient
from fastapi_simple_cache import FastAPISimpleCache
from fastapi_simple_cache.decorator import cache
from fastapi_simple_cache.backends.inmemory import InMemoryBackend
from fastapi_simple_cache.backends.firestore import FirestoreBackend
from fastapi_simple_cache.backends.redis import RedisBackend

from .backends.firestore import CollectionReference
from .backends.redis import Redis

app = FastAPI()
client = TestClient(app)


@app.post("/")
@cache(expire=2)
def root(request: Request):
    return {"epoch": time.time()}


@pytest.fixture
def firestore_backend():
    backend = FirestoreBackend(collection=CollectionReference())
    FastAPISimpleCache.reset()
    FastAPISimpleCache.init(backend=backend)


@pytest.fixture
def redis_backend():
    backend = RedisBackend(redis=Redis())
    FastAPISimpleCache.reset()
    FastAPISimpleCache.init(backend=backend)


@pytest.fixture
def multi_backend():
    backend_inmem = InMemoryBackend()
    backend_redis = RedisBackend(redis=Redis())
    FastAPISimpleCache.reset()
    FastAPISimpleCache.init(backend=[backend_inmem, backend_redis])


def test_firestore(firestore_backend):
    response = client.post("/")
    epoch = response.json().get("epoch")
    assert response.headers.get("age") == "0"
    time.sleep(1)
    response = client.post("/")
    assert response.json().get("epoch") == epoch
    assert response.headers.get("age") == "1"
    time.sleep(1)
    response = client.post("/")
    assert response.json().get("epoch") != epoch
    assert response.headers.get("age") == "0"


def test_redis(redis_backend):
    response = client.post("/")
    epoch = response.json().get("epoch")
    assert response.headers.get("age") == "0"
    time.sleep(1)
    response = client.post("/")
    assert response.json().get("epoch") == epoch
    assert response.headers.get("age") == "1"
    time.sleep(1)
    response = client.post("/")
    assert response.json().get("epoch") != epoch
    assert response.headers.get("age") == "0"


def test_multi_backend(multi_backend, caplog):
    response = client.post("/")
    epoch = response.json().get("epoch")
    assert response.headers.get("age") == "0"
    time.sleep(1)
    response = client.post("/")
    assert response.json().get("epoch") == epoch
    assert response.headers.get("age") == "1"
    time.sleep(1)
    response = client.post("/")
    assert response.json().get("epoch") != epoch
    assert response.headers.get("age") == "0"
    assert re.match(r"Set .*? to InMemoryBackend", caplog.records[0].message)
    assert re.match(r"Set .*? to RedisBackend", caplog.records[1].message)
    assert re.match(r"Get .*? from InMemoryBackend", caplog.records[2].message)
    assert re.match(r"Exp .*? from InMemoryBackend", caplog.records[3].message)
    assert re.match(r"Set .*? to InMemoryBackend", caplog.records[4].message)
    assert re.match(r"Set .*? to RedisBackend", caplog.records[5].message)



from datetime import datetime
from fastapi import FastAPI, Request
from fastapi_simple_cache import FastAPISimpleCache
from fastapi_simple_cache.backends.redis import RedisBackend
from fastapi_simple_cache.decorator import cache
from redis.asyncio import ConnectionPool, client

app = FastAPI()


@app.on_event("startup")
async def startup():
    pool = ConnectionPool.from_url(url="redis://localhost:6379")
    backend = RedisBackend(redis=client.Redis(connection_pool=pool))
    FastAPISimpleCache.init(backend=backend)


@app.get("/")
@cache(expire=10)
def root(request: Request):
    return {"datetime": datetime.utcnow()}



import json
from datetime import datetime, timedelta
from typing import Mapping, Optional


class Pipeline:
    def __init__(
        self,
        redis,
    ):
        self.redis = redis
        self.key = None

    async def __aenter__(self):
        return self

    async def __aexit__(self, *args, **kwargs):
        self.key = None

    def ttl(self, key):
        return self

    def get(self, key):
        self.key = key
        return self

    async def execute(self):
        value = self.redis.data.get(self.key)
        if value is None:
            return None, None
        data = json.loads(value)
        expire_at = datetime.fromisoformat(data.get("expire_at"))
        ttl = (expire_at - datetime.utcnow()).total_seconds()
        return ttl, value


class Redis:
    def __init__(
        self,
        data: Optional[Mapping] = None,
    ):
        if data is None:
            data = {}
        self.data = data

    def pipeline(
        self,
        **kwargs,
    ):
        return Pipeline(redis=self)

    async def set(self, key: str, value: str, ex: int):
        data = json.loads(value)
        data["expire_at"] = (datetime.utcnow() + timedelta(seconds=ex)).isoformat()
        value = json.dumps(data)
        self.data[key] = value



"""FastAPI Simple Cache"""

__version__ = "0.1.1"

import asyncio
import logging
from fastapi import Response
from typing import List, Optional, Tuple, Union

from .backends import Backend

logger = logging.getLogger(__name__)


class FastAPISimpleCache:
    backends = None
    namespace = None

    @classmethod
    def init(
        cls,
        backend: Union[Backend, List[Backend]],
        namespace: Optional[str] = None,
    ) -> None:
        if isinstance(backend, Backend):
            backend = [backend]
        cls.backends = backend
        cls.namespace = namespace
        pass

    @classmethod
    async def get(
        cls,
        key: str,
    ) -> Tuple[Optional[Response], Optional[int]]:
        cls._check_init()
        for backend in cls.backends:
            response, ttl = await backend.get(key)
            if response:
                return response, ttl
        return None, None

    @classmethod
    def set(
        cls,
        key: str,
        response: Response,
        expire: int,
    ) -> None:
        cls._check_init()
        loop = asyncio.get_running_loop()
        for backend in cls.backends:
            loop.create_task(backend.set(key, response, expire))
        pass

    @classmethod
    def _check_init(cls):
        if cls.backends is None:
            raise Exception("You must call init first")
        pass

    @classmethod
    def reset(cls):
        cls.backends = None
        cls.namespace = None
        pass
